""" Helper function for getting things in and out of the database """

import os
from collections import namedtuple, ChainMap
import itertools
import subprocess
import argparse

import cv2
import zbar
import numpy as np
import pandas

from pony import orm
from pony.orm import Database

from .models import Student, Grader, Exam, Submission, \
                    Page, Problem, FeedbackOption, Solution

### Database definition.

db = Database()
session = orm.db_session


### Auxiliary functions dealing with files and images

ExtractedQR = namedtuple('ExtractedQR', ['name', 'page', 'sub_nr', 'coords'])


def pdf_to_images(filename):
    """Extract all images out of a pdf file."""
    # We convert everything to jpeg, which may be suboptimal, however some
    # formats recognized by pdfimages aren't understood by opencv.
    subprocess.run(['pdfimages', '-j', filename, filename[:-len('.pdf')]])


def guess_dpi(image_array):
    h, *_ = image_array.shape
    resolutions = np.array([1200, 600, 300, 200, 150, 120, 100, 75, 60, 50, 40])
    return resolutions[np.argmin(abs(resolutions - 25.4 * h / 297))]


def get_box(image_array, box, padding):
    """Extract a subblock from an array corresponding to a scanned A4 page.

    Parameters:
    -----------
    image_array : 2D or 3D array
        The image source.
    box : 4 floats (top, bottom, left, right)
        Coordinates of the bounding box in inches. By due to differing
        traditions, box coordinates are counted from the bottom left of the
        image, while image array coordinates are from the top left.
    padding : float
        Padding around box borders in inches.
    """
    h, w, *_ = image_array.shape
    dpi = guess_dpi(image_array)
    box = np.array(box)
    box += (padding, -padding, -padding, padding)
    box = (np.array(box) * dpi).astype(int)
    # Here we are not returning the lowest pixel of the image because otherwise
    # the numpy slicing is not correct.
    top, bottom = min(h, box[0]), max(1, box[1])
    left, right = max(0, box[2]), min(w, box[3])
    return image_array[-top:-bottom, left:right]


def extract_qr(image_path, yaml_version, scale_factor=4):
    image = cv2.imread(image_path,
                       cv2.IMREAD_GRAYSCALE)[::scale_factor, ::scale_factor]
    if image.shape[0] < image.shape[1]:
        image = image.T
    # Varied thresholds because zbar is picky about contrast.
    for threshold in (200, 150, 220):
        thresholded = 255 * (image > threshold)
        # zbar also cares about orientation.
        for direction in itertools.product([1, -1], [1, -1]):
            flipped = thresholded[::direction[0], ::direction[1]]
            scanner = zbar.Scanner()
            results = scanner.scan(flipped.astype(np.uint8))
            if results:
                try:
                    version, name, page, copy = \
                                results[0].data.decode().split(';')
                except ValueError:
                    return
                if version != 'v{}'.format(yaml_version):
                    raise RuntimeError('Yaml format mismatch')
                coords = np.array(results[0].position)
                # zbar doesn't respect array ordering!
                if not np.isfortran(flipped):
                    coords = coords[:, ::-1]
                coords *= direction
                coords %= image.shape
                coords *= scale_factor
                return ExtractedQR(name, int(page), int(copy), coords)
    else:
        return


def rotate_and_shift(image_path, extracted_qr, qr_coords):
    _, page, _, position = extracted_qr
    image = cv2.imread(image_path)

    if image.shape[0] < image.shape[1]:
        image = np.transpose(image, (1, 0, 2))

    dpi = guess_dpi(image)
    h, w, *_ = image.shape
    qr_widget = qr_coords[qr_coords.page == page]
    box = dpi * qr_widget[['top', 'bottom', 'left', 'right']].values[0]
    y0, x0 = h - np.mean(box[:2]), np.mean(box[2:])
    y, x = np.mean(position, axis=0)
    if (x > w / 2) != (x0 > w / 2):
        image = image[:, ::-1]
        x = w - x
    if (y > h / 2) != (y0 > h / 2):
        image = image[::-1]
        y = h - y

    shift = np.round((y0-y, x0-x)).astype(int)
    shifted_image = np.roll(image, shift[0], axis=0)
    shifted_image = np.roll(shifted_image, shift[1], axis=1)

    cv2.imwrite(image_path, shifted_image)


def get_widget_image(image_path, widget):
    box = (widget.top, widget.bottom, widget.left, widget.right)
    raw_image = get_box(cv2.imread(image_path), box, padding=0.3)
    return cv2.imencode(".jpg", raw_image)[1].tostring()


def solution_data(exam_id, student_id):
    """Return Python datastructures corresponding to the student submission."""
    with orm.db_session:
        exam = Exam[exam_id]
        student = Student[student_id]
        if any(i is None for i in (exam, student)):
            raise RuntimeError('Student did not make a '
                               'submission for this exam')

        results = []
        for problem in exam.problems.order_by(Problem.id):
            if not orm.count(problem.solutions.feedback):
                # Nobody received any grade for this problem
                continue
            problem_data = {
                'name': problem.name,
                'max_score': orm.max(problem.feedback_options.score, default=0)
            }
            solutions = Solution.select(lambda s: s.problem == problem
                                        and s.submission.student == student)
            problem_data['feedback'] = [
                {'short': fo.text,
                 'score': fo.score,
                 'description': fo.description}
                for solution in solutions for fo in solution.feedback
            ]
            problem_data['score'] = sum(i['score'] or 0
                                        for i in problem_data['feedback'])
            problem_data['remarks'] = '\n\n'.join(sol.remarks
                                                  for sol in solutions
                                                  if sol.remarks)
            results.append(problem_data)

        pages = Page.select(lambda p: p.submission.exam == exam
                            and p.submission.student == student)
        # 'set' invokation is a hotfix for issue #50
        paths = sorted(set(page.path for page in pages))

        student = student.to_dict()

    student['total'] = sum(i['score'] for i in results)
    return student, results, paths


def full_exam_data(exam_id):
    """Export all grades of an exam as a pandas DataFrame."""
    with orm.db_session:
        students = sorted(Exam[exam_id].submissions.student.id)

        data = [solution_data(exam_id, student_id)
                for student_id in students]

    students = pandas.DataFrame({i[0]['id']: i[0] for i in data}).T
    del students['id']

    results = {}
    for result in data:
        for problem in result[1]:
            name = problem.pop('name')
            problem[(name, 'remarks')] = problem.pop('remarks')
            for fo in problem.pop('feedback'):
                problem[(name, fo['short'])] = fo['score']
            problem[(name, 'total')] = problem.pop('score')
            problem.pop('max_score')
        results[result[0]['id']] = dict(ChainMap({('total', 'total'):
                                                  result[0]['total']},
                                                 *result[1]))

    return pandas.DataFrame(results).T


def get_student_number(image_path, widget):
    """Extract the student number from the image path with the scanned number.

    TODO: all the numerical parameters are guessed and work well on the first
    exam.  Yet, they should be inferred from the scans.
    """
    image = cv2.imread(image_path, cv2.IMREAD_GRAYSCALE)
    box = (widget.top, widget.bottom, widget.left, widget.right)
    image = get_box(image, box, padding=0.3)
    _, thresholded = cv2.threshold(image, 150, 255, cv2.THRESH_BINARY)
    thresholded = cv2.bitwise_not(thresholded)

    # Copy the thresholded image.
    im_floodfill = thresholded.copy()

    # Mask used to flood filling.
    # Notice the size needs to be 2 pixels than the image.
    h, w, *_ = thresholded.shape
    mask = np.zeros((h+2, w+2), np.uint8)

    # Floodfill from point (0, 0)
    cv2.floodFill(im_floodfill, mask, (0,0), 255);

    # Invert floodfilled image
    im_floodfill_inv = cv2.bitwise_not(im_floodfill)

    # Combine the two images to get the foreground.
    im_out = ~(thresholded | im_floodfill_inv)

    params = cv2.SimpleBlobDetector_Params()
    params.filterByArea = True
    params.minArea = 500
    params.maxArea = 1500
    params.filterByCircularity = True
    params.minCircularity = 0.01
    params.filterByConvexity = True
    params.minConvexity = 0.87
    params.filterByInertia = True
    params.minInertiaRatio = 0.7

    detector = cv2.SimpleBlobDetector_create(params)

    keypoints = detector.detect(im_out)
    centers = np.array(sorted([kp.pt for kp in keypoints])).astype(int)
    right, bottom = np.max(centers, axis=0)
    left, top = np.min(centers, axis=0)
    centers = np.mgrid[left:right:10j, top:bottom:7j].astype(int)
    weights = []
    for center in centers.reshape(2, -1).T:
        x, y = center
        r = 12
        weights.append(np.sum(255 - image[y-r:y+r, x-r:x+r]))
    weights = np.array(weights).reshape(10, 7)
    return sum(((np.argmax(weights, axis=0) + 1) % 10)[::-1] * 10**np.arange(7))

### Combining everything: build the database.

def init_db(db_file='course.sqlite', overwrite=False):
    create_db = False
    if not os.path.exists(db_file) or overwrite:
        try:
            os.remove(db_file)
        except OSError:
            pass
        create_db = True

    try:
        db.bind('sqlite', db_file, create_db=create_db)
        db.generate_mapping(create_tables=True)
    except TypeError as exc:  # can raise if db is already bound
        if 'already bound' not in str(exc):
            raise


def add_participants(students='students.csv', graders='graders.csv'):
    init_db()
    with db_session:
        students = pandas.read_csv(students, skipinitialspace=True,
                                   header=None)
        for student_id, first_name, last_name, email in students.values:
            try:
                Student(first_name=first_name, last_name=last_name,
                        id=student_id, email=(email if not pandas.isnull(email)
                                              else None))
            except orm.CacheIndexError:
                pass

        graders = pandas.read_csv(graders, skipinitialspace=True, header=None)
        for first_name, last_name in graders.values:
            try:
                Grader(first_name=first_name, last_name=last_name)
            except orm.CacheIndexError:
                pass


def add_exam(yaml_path):
    init_db()
    _, exam_name, qr_coords, widget_data = read_yaml(yaml_path)
    data_dir = exam_name + '_data'
    os.makedirs(data_dir, exist_ok=True)
    with orm.db_session:
        # init exam
        exam = Exam.get(name=exam_name) or Exam(name=exam_name,
                                                yaml_path=yaml_path)

        # Default feedback (maybe factor out eventually).
        feedback_options = ['Everything correct',
                            'No solution provided']

        for name in widget_data.index:
            if name == 'studentnr':
                continue
            p = Problem(name=name, exam=exam)
            for fb in feedback_options:
                FeedbackOption(text=fb, problem=p)


def update_exam(exam_name, yaml_path):
    _, exam_name, _, widgets = read_yaml(yaml_path)
    new_problem_names = list(name for name in widgets.index
                             if name != 'studentnr')

    with orm.db_session:
        exam = Exam.get(name=exam_name)
        exam.yaml_path = yaml_path
        problems = list(Problem.select(lambda p: p.exam == exam)
                               .order_by(lambda p: p.id))
        for problem, name in zip(problems, new_problem_names):
            problem.name = name


def update_problem_names(exam_name, new_problem_names):
    with orm.db_session:
        exam = Exam.get(name=exam_name)
        problems = list(Problem.select(lambda p: p.exam == exam)
                               .order_by(lambda p: p.id))
        if len(new_problem_names) != len(problems):
            raise ValueError('Number of existing problems and new problem '
                             'names differ.')
        for problem, name in zip(problems, new_problem_names):
            problem.name = str(name)


def process_pdf(pdf_path, meta_yaml, verify_name=True):
    """Add all information from a pdf to the app (database + file storage).

    Returns the indices of pages that couldn't be processed.
    Parameters:
    -----------
    verify_name : Bool
        Whether to check that the uploaded pdf belongs to the correct exam.
        Useful to accommodate a common user error.
    """
    init_db()
    # read in metadata
    version, exam_name, qr_coords, widget_data = read_yaml(meta_yaml)
    with orm.db_session:
        exam = Exam.get(name=exam_name)
        if not exam:
            raise RuntimeError('Exam {} does not exist in the database'
                               .format(exam_name))
    # create images
    source_dir, pdf_filename = os.path.split(pdf_path)
    ## shutil.copy(pdf_path, source_dir)
    pdf_to_images(os.path.join(source_dir, pdf_filename))
    images = filter((lambda name: (not name.endswith('.pdf'))
                                  and name.startswith(pdf_filename[:-4])),
                    os.listdir(source_dir))
    images =  sorted(os.path.join(source_dir, image) for image in images)
    # verify that copies are from the same exam
    extracted_qrs = [extract_qr(image, version) for image in images]
    if not verify_name and any(qr[0] != exam_name for qr in extracted_qrs
                               if qr is not None):
        for image in images:
            os.remove(image)
        raise RuntimeError('metadata and pdf are from different exams')

    no_qrs = []
    for image, qr in zip(images, extracted_qrs):
        if qr is None:
            # Extract the page number using the naming scheme of pdfimages.
            no_qrs.append(int(image[image.rfind('-') + 1 : image.rfind('.')]) +
                          1)
        else:
            rotate_and_shift(image, qr, qr_coords)

    for image, qr_data in zip(images, extracted_qrs):
        if qr_data is None:
            continue
        sub_nr = qr_data.sub_nr
        with orm.db_session:
            exam = Exam.get(name=exam_name)
            sub = Submission.get(copy_number=sub_nr, exam=exam) \
                  or Submission(copy_number=sub_nr, exam=exam)
            extension = image[image.rfind('.'):]
            base = os.path.split(image)[0]
            target = os.path.join(base, qr_data.name + '_{}'.format(sub_nr))
            os.makedirs(target, exist_ok=True)
            target_image = os.path.join(target, 'page' + str(qr_data.page)
                                                + extension)
            os.rename(image, target_image)
            # We may have added this page in previous uploads; the above
            # 'rename' then overwrites the previosly uploaded page, but
            # we only want a single 'Page' entry.
            if Page.get(path=target_image, submission=sub) is None:
                Page(path=target_image, submission=sub)
            widgets_on_page = widget_data[widget_data.page == qr_data.page]
            for problem in widgets_on_page.index:
                if problem == 'studentnr':
                    sub.signature_image_path = 'None'
                    try:
                        number_widget = widgets_on_page.loc['studentnr']
                        number = get_student_number(target_image, number_widget)
                        sub.student = Student.get(id=int(number))
                    except Exception:
                        pass  # could not extract student name
                else:
                    prob = Problem.get(name=problem, exam=exam)
                    sol = Solution.get(problem=prob, submission=sub)
                    if sol:
                        sol.image_path = 'None'
                    else:
                        Solution(problem=prob, submission=sub,
                                 image_path='None')

    return no_qrs


def do_everything(scanned_pdf, meta_yaml, students='students.csv',
                  graders='graders.csv', overwrite=False):
    init_db(overwrite=overwrite)
    add_participants(students, graders)
    add_exam(meta_yaml)
    process_pdf(scanned_pdf, meta_yaml)


def main():
    parser = argparse.ArgumentParser(description='Create a new exam '
                                     'for grading.')
    parser.add_argument('--overwrite', action='store_true',
                        help='destroy the existing database')
    parser.add_argument('pdf', help='Scanned exam pdf')
    parser.add_argument('yaml', help='Meta-information about the exam')
    args = parser.parse_args()
    if args.overwrite:
        rsp = input('WARNING: we are going to wipe the database, is this ok? ')
        if not rsp.lower().startswith('y'):
            print('Quitting without overwriting the database')
            exit(0)
        else:
            print('Carrying on...')
    do_everything(args.pdf, args.yaml, overwrite=args.overwrite)

    with orm.db_session:
        print('Graders\n' + '-' * 7)
        Grader.select().show()
        print()
        print('Students\n' + '-' * 8)
        Student.select().show()


def use_db(filename='course.sqlite'):
    db.bind('sqlite', filename)
    db.generate_mapping(create_tables=True)


if __name__ == '__main__':
    main()



from itertools import starmap
import os
import datetime

import traitlets
from pony import orm

import db


def submission_to_key(sub):
    stud = None if sub.student is None else sub.student.id
    return (stud, sub.id)


class AppModel(traitlets.HasTraits):
    # Immutable
    students = traitlets.List(traitlets.Unicode(), read_only=True)
    graders = traitlets.List(read_only=True)
    exams = traitlets.List(read_only=True)

    exam_id = traitlets.Integer()
    problems = traitlets.List()

    submission_id = traitlets.Integer()
    student = traitlets.Unicode()

    # Stack storing which submissions we have seen, used for working
    # the "prev" button.
    seen_submissions = []

    # Default grader_id = 0 is definitely not in the database.
    # Here we are abusing 1-based indexing and the internal db structure.
    grader_id = traitlets.Integer()

    problem_id = traitlets.Integer()
    n_solutions = traitlets.Integer()
    n_graded = traitlets.Integer()
    feedback_options = traitlets.List()

    selected_feedback = traitlets.List(trait=traitlets.Integer())
    remarks = traitlets.Unicode()

    show_full_page = traitlets.Bool(default_value=False)

    edited_feedback_option = traitlets.Integer()
    edited_feedback_name = traitlets.Unicode()
    edited_score = traitlets.Integer()
    edited_description = traitlets.Unicode()

    # --- Defaults and validation ---
    ## students
    @traitlets.default('students')
    def _default_students(self):
        with orm.db_session:
            return tuple(starmap('{} ({} {})'.format,
                                 orm.select((s.id, s.first_name, s.last_name)
                                            for s in db.Student)))

    ## graders
    @traitlets.default('graders')
    def _default_graders(self):
        with orm.db_session:
            graders = [(f'{g.first_name} {g.last_name}', g.id)
                       for g in orm.select(g for g in db.Grader)
                                   .order_by(lambda g: g.id)]
        return [("None", 0)] + graders

    ## problems
    @traitlets.default('problems')
    def _default_problems(self):
        with orm.db_session:
            problems = (orm.select(p for p in db.Problem
                                   if p.exam.id == self.exam_id)
                           .order_by(lambda p: p.id))
            return list((p.name, p.id) for p in problems)

    ## exams
    @traitlets.default('exams')
    def _default_exams(self):
        with orm.db_session:
            exams = orm.select(e for e in db.Exam).order_by(lambda e: e.id)
            return [(e.name, e.id) for e in exams]

    ## exam_id
    @traitlets.default('exam_id')
    def _default_exam_id(self):
        with orm.db_session:
            # Nonempty exam with the most recent yaml
            return sorted(db.Exam.select(lambda e: len(e.submissions)),
                          key=lambda e: os.path.getmtime(e.yaml_path))[-1].id

    @traitlets.validate('exam_id')
    def _valid_exam(self, proposal):
        with orm.db_session():
            if db.Exam.get(id=proposal['value']) is None:
                raise traitlets.TraitError('Unknown exam id.')
        return proposal['value']

    ## submission_id
    @traitlets.default('submission_id')
    def _default_submission_id(self):
        with orm.db_session:
            return (orm.select(s.id for s in db.Submission
                              if s.exam.id == self.exam_id)
                       .order_by(lambda x:x)
                       .first())

    @traitlets.validate('submission_id')
    def _valid_submission_id(self, proposal):
        with orm.db_session():
            sub = db.Submission.get(id=proposal['value'])
            if sub is None:
                raise traitlets.TraitError('Unknown submission id.')
            if sub.exam.id != self.exam_id:
                raise traitlets.TraitError('Submission from a different exam.')
        return proposal['value']

    ## problem_id
    @traitlets.default('problem_id')
    def _default_problem_id(self):
        with orm.db_session:
            return (orm.select(p.id for p in db.Problem
                               if p.exam == db.Exam[self.exam_id])
                    .order_by(lambda x:x).first())

    @traitlets.validate('problem_id')
    def _valid_problem_id(self, proposal):
        with orm.db_session():
            prob = db.Problem.get(id=proposal['value'])
            if prob is None:
                raise traitlets.TraitError('Unknown problem id.')
            if prob.exam.id != self.exam_id:
                raise traitlets.TraitError('Problem from a different exam.')
        return proposal['value']

    ## n_solutions
    @traitlets.default('n_solutions')
    def _default_n_solutions(self):
        with orm.db_session:
            return orm.count(s for s in db.Problem[self.problem_id].solutions)

    ## n_graded
    @traitlets.default('n_graded')
    def _default_n_graded(self):
        with orm.db_session:
            return orm.count(s for s in db.Problem[self.problem_id].solutions
                             if s.graded_at is not None)

    ## grader_id
    @traitlets.validate('grader_id')
    def _valid_grader_id(self, proposal):
        new = proposal['value']
        if new == 0:
            return new
        else:
            with orm.db_session:
                if db.Grader.get(id=new) is None:
                    raise traitlets.TraitError('Non-existent grader id.')
        return new

    ## feedback_options
    @traitlets.default('feedback_options')
    def _default_feedback_options(self):
        def formatted(fo):
            text = fo.text + f' ({fo.score})' * (fo.score is not None)
            return text, fo.id

        with orm.db_session:
            fo = (db.Problem[self.problem_id]
                    .feedback_options
                    .order_by(lambda f: f.id))
            return list(formatted(f) for f in fo)

    ## selected_feedback
    @traitlets.default('selected_feedback')
    def _default_selected_feedback(self):
        with orm.db_session:
            sol = db.Solution.get(submission=self.submission_id,
                                  problem=self.problem_id)
            if sol is None:
                return []
            return [f.id for f in sol.feedback]

    ## remarks
    @traitlets.default('remarks')
    def _default_remarks(self):
        with orm.db_session:
            sol = db.Solution.get(submission=self.submission_id,
                                  problem=self.problem_id)
            if sol is None:
                return ''
            return sol.remarks or ''

    ## student
    @traitlets.default('student')
    def _default_student(self):
        with orm.db_session:
            student = db.Submission[self.submission_id].student
            if student is None:
                return "MISSING"
            else:
                return f"{student.id} ({student.first_name} {student.last_name})"

    # --- Relations between traits ---
    @traitlets.observe('exam_id')
    def _change_exam(self, change):
        self.submission_id = self._default_submission_id()
        self.problems = self._default_problems()
        self.problem_id = self._default_problem_id()

    @traitlets.observe('problem_id', 'submission_id')
    def _change_solution(self, change):
        self.commit_grading(**{change['name']: change['old']})
        self.student = self._default_student()
        self.remarks = self._default_remarks()
        self.feedback_options = []  # hack to fix bug in ipywidgets
        self.feedback_options = self._default_feedback_options()
        self.selected_feedback = self._default_selected_feedback()

    @traitlets.observe('edited_feedback_option')
    def _update_edited_option(self, change):
        if self.edited_feedback_option:
            with orm.db_session:
                fo = db.FeedbackOption[self.edited_feedback_option]
                assert fo.problem.id == self.problem_id
                self.edited_feedback_name = fo.text or ''
                self.edited_description = fo.description or ''
                self.edited_score = fo.score or 0
        else:
            self.edited_feedback_name = ''
            self.edited_score = 0
            self.edited_description = ''

    # --- Writing into database ---
    def commit_grading(self, submission_id=None, problem_id=None):
        """Commit grading to the database"""
        if submission_id is None:
            submission_id = self.submission_id
        if problem_id is None:
            problem_id = self.problem_id

        # Sometimes we transition from a nonexistent solution and end up here.
        if db.Solution.get(submission=submission_id,
                           problem=problem_id) is None:
            return

        # Should do nothing when the grader is missing.
        if not self.grader_id:
            return

        with orm.db_session:
            solution = db.Solution.get(submission=submission_id,
                                       problem=problem_id)
            old_feedback = set(fb.id for fb in solution.feedback)
            old_remarks = solution.remarks
            # Check if anything changed -- if not don't save anything
            if (set(self.selected_feedback) == old_feedback
                    and self.remarks.strip() == old_remarks):
                return

            # Otherwise, save everything
            solution.feedback = [db.FeedbackOption[i] for i
                                 in self.selected_feedback]
            solution.remarks = self.remarks
            if len(solution.feedback):
                solution.graded_by = db.Grader[self.grader_id]
                solution.graded_at = datetime.datetime.now()
            else:  # if no feedback, then the problem is not graded
                solution.graded_by = None
                solution.graded_at = None

    def commit_feedback_edit(self, *_):
        self.edited_feedback_name = self.edited_feedback_name.strip()
        if not self.edited_feedback_name:
            return
        with orm.db_session:
            # Check if we create a collision.
            # TODO: reflect a noop in UI via a Valid widget
            if self.edited_feedback_option:
                orig_text = db.FeedbackOption[self.edited_feedback_option].text
            else:
                orig_text = None
            if (self.edited_feedback_name != orig_text
                    and db.FeedbackOption.get(text=self.edited_feedback_name,
                                              problem=self.problem_id)
                        is not None):
                self.edited_feedback_option = 0
                return
        description = self.edited_description.strip()
        score = self.edited_score

        if not self.edited_feedback_option:
            with orm.db_session:
                fo = db.FeedbackOption(text=self.edited_feedback_name,
                                       problem=db.Problem[self.problem_id])
            option = fo.id
        else:
            option = self.edited_feedback_option

        # Set the details in the database.
        with orm.db_session:
            fo = db.FeedbackOption[option]
            fo.text = self.edited_feedback_name.strip()
            fo.description = description
            fo.score = score

        self.feedback_options = []  # hack to fix bug in ipywidgets
        self.feedback_options = self._default_feedback_options()
        self.selected_feedback = self._default_selected_feedback()
        self.edited_feedback_option = option

    def delete_feedback_option(self, *_):
        value = self.edited_feedback_option
        if not value:
            return

        with orm.db_session:
            db.FeedbackOption[value].delete()
            # Some solutions may become ungraded.
            # We should reflect that in the UI
            for solution in db.Solution.select(lambda s:
                                               s.problem.id == self.problem_id
                                               and not s.feedback):
                solution.graded_by = solution.graded_at = None

        self.feedback_options = []  # hack to fix bug in ipywidgets
        self.feedback_options = self._default_feedback_options()
        self.selected_feedback = self._default_selected_feedback()

    # --- Navigation
    def next_submission(self):
        with orm.db_session:
            own_sub = db.Submission[self.submission_id]
            own_key = (own_sub, own_sub.student)
            # Note the "wrong" ordering in the second filter.
            # It appears, there's a bug in pony that translates tuple
            # comparison incorrectly
            subs = (db.Submission.select()
                      .filter(lambda s: s.exam.id == self.exam_id)
                      .filter(lambda s: (s, s.student) > own_key)
                      .order_by(lambda s: (s.student, s)))
            result = subs.first()
            if result is None:
                subs = (db.Submission.select()
                          .filter(lambda s: s.exam.id == self.exam_id)
                          .order_by(lambda s: (s.student, s)))
                result = subs.first()

            self.seen_submissions.append(self.submission_id)
            self.submission_id = result.id

    def previous_submission(self):
        try:
            self.submission_id = self.seen_submissions.pop()
        except IndexError:
            pass  # don't move if we have not seen any submissions yet

    def next_ungraded(self):
        with orm.db_session:
            all_ungraded = orm.select(s.submission.id for s in db.Solution
                                      if s.submission.exam.id == self.exam_id
                                      and s.problem.id == self.problem_id
                                      and s.graded_at is None).order_by(lambda
                                                                        x: x)[:]

            for filt in (lambda s: s > self.submission_id, lambda s: True):
                try:
                    next_sub_id = next(filter(filt, all_ungraded))
                    self.seen_submissions.append(self.submission_id)
                    self.submission_id = next_sub_id
                    return
                except StopIteration:
                    continue

    def jump_to_student(self, nr):
        with orm.db_session:
            sub = db.Submission.select(lambda s: s.exam.id==self.exam_id
                                               and s.student.id==nr).first()
            sub_id = sub.id if sub else None
        if sub_id is None:
            return

        self.seen_submissions.append(self.submission_id)
        self.submission_id = sub_id

    # --- Other methods
    def num_submissions(self, exam_id):
        with orm.db_session:
            return orm.count(s for s in db.Submission if s.exam.id == exam_id)

    def num_graded(self):
        with orm.db_session:
            this_problem = db.Problem[self.problem_id]
            return sum(bool(s.graded_by) for s in this_problem.solutions)

    def get_solution(self):
        with orm.db_session:
            p = db.Problem[self.problem_id]
            s = db.Solution.get(submission=self.submission_id,
                                problem=self.problem_id)

            graded_at = grader_name = None
            image = b''
            if s:
                *_, widgets = self.exam_metadata()
                problem_metadata = widgets.loc[p.name]
                # the '.' at the end is important, otherwise this would
                # match 'page12' when we were meant to be looking for 'page1'
                page = f'page{int(problem_metadata.page)}.'
                # Here we use the specific page naming scheme because the
                # database does not store the page order.
                # Eventually the database should be restructured to make
                # this easier.
                page_image_path = (s.submission.pages
                                    .select(lambda p: page in p.path)
                                    .first().path)
                if self.show_full_page:
                    with open(page_image_path, 'rb') as f:
                        image = f.read()
                elif s.image_path != 'None':
                    with open(s.image_path, 'rb') as f:
                        image = f.read()
                else:
                    image = db.get_widget_image(page_image_path,
                                                problem_metadata)

                graded_at = s.graded_at

                if s.graded_by:
                    grader_name =  (f'{s.graded_by.first_name} '
                                    f'{s.graded_by.last_name}')
            return image, (grader_name, graded_at)

    def exam_metadata(self):
        with orm.db_session:
            fname = db.Exam[self.exam_id].yaml_path
        return db.read_yaml(fname)



from flask import Blueprint
from flask_restful import Api

from .resources.graders import Graders
from .resources.exams import Exams, ExamConfig

api_bp = Blueprint(__name__, __name__)

errors = {
    'ObjectNotFound': {
        'status': 404,
        'message': 'Resource with that ID does not exist',
     },
}

api = Api(api_bp, errors=errors)

api.add_resource(Graders, '/graders')
api.add_resource(Exams, '/exams')
api.add_resource(ExamConfig, '/exams/<int:exam_id>')



""" Init file that starts a Flask dev server and opens db """

import os
from os.path import abspath, dirname

from flask import Flask
from werkzeug.exceptions import NotFound
from flask_basicauth import BasicAuth

from .api import api_bp
from .models import db


STATIC_FOLDER_PATH = os.path.join(abspath(dirname(__file__)), 'static')

app = Flask(__name__, static_folder=STATIC_FOLDER_PATH)
app.register_blueprint(api_bp, url_prefix='/api')
auth = BasicAuth()  # don't pass 'app', as its not yet configured

# Default settings
app.config.update(
    DATA_DIRECTORY=os.path.join(os.getcwd(), 'data')
)
if 'ZESJE_SETTINGS' in os.environ:
    app.config.from_envvar('ZESJE_SETTINGS')


@app.before_first_request
def setup():
    auth.init_app(app)

    data_dir = app.config['DATA_DIRECTORY']
    if not os.path.exists(data_dir):
        os.makedirs(data_dir)

    db.bind('sqlite', f"{app.config['DATA_DIRECTORY']}/course.sqlite", create_db=True)
    db.generate_mapping(create_tables=True)


@app.route('/')
@app.route('/<file>')
def index(file='index.html'):
    """Serve the static react content, otherwise fallback to the index.html"""
    try:
        return app.send_static_file(file)
    except NotFound:
        return app.send_static_file('index.html')



from flask import Blueprint, jsonify, request, redirect

from . import db

app = Blueprint(__name__, __name__)

@app.route('/students', methods=['GET'])
@db.session
def get_students():
    return jsonify([
        dict(id=s.id, first_name=s.first_name,
             last_name=s.last_name, email=s.email)
        for s in db.Student.select()
    ])


@app.route('/students', methods=['POST'])
@db.session
def upload_students():
    if 'file' not in request.files:
        flash('No file part')
        return redirect(request.url)
    file = request.files['file']
    # if user does not select file, browser also
    # submit a empty part without filename
    if file.filename == '':
        flash('No selected file')
        return redirect(request.url)
    if file and allowed_file(file.filename):
        filename = secure_filename(file.filename)
        file.save(os.path.join(app.config['UPLOAD_FOLDER'], filename))
        return redirect(url_for('uploaded_file',
                                filename=filename))



from setuptools import setup, find_packages
from distutils.command.build import build
from setuptools.command.sdist import sdist


def webpack():
    import subprocess
    subprocess.check_call(['yarn', 'install'])
    subprocess.check_call(['yarn', 'build'])


class Build(build):
    def run(self):
        webpack()
        super().run()


class Sdist(sdist):
    def run(self):
        webpack()
        super().run()


setup(
    name="zesje",
    version="0.1a",
    url="http://gitlab.kwant-project,org/zesje/zesje",
    description="",
    author="Zesje authors",
    author_email="anton.akhmerov@tudelft.nl",
    packages=find_packages('.'),
    cmdclass={'build': Build,
              'sdist': Sdist,
             },
    package_data={'zesje': ['static/*']},
    include_package_data=True,
)



""" Models used in the db """
from pony.orm import Database, Required, Optional, PrimaryKey, Set

from datetime import datetime

db = Database()


class Student(db.Entity):
    """ New students may be added throughout the course. """
    id = PrimaryKey(int)
    first_name = Required(str)
    last_name = Required(str)
    email = Optional(str, unique=True)
    submissions = Set('Submission')


class Grader(db.Entity):
    """ This will be initialized @ app initialization and immutable from then on. """
    first_name = Required(str)
    last_name = Required(str)
    graded_solutions = Set('Solution')


class Exam(db.Entity):
    """ New instances are created when providing a new exam. """
    name = Required(str, unique=True)
    yaml_path = Required(str)
    submissions = Set('Submission')
    problems = Set('Problem')


class Submission(db.Entity):
    """ Typically created when adding a new exam. """
    copy_number = Required(int)
    exam = Required(Exam)
    solutions = Set('Solution')
    pages = Set('Page')
    student = Optional(Student)
    signature_image_path = Optional(str)
    signature_validated = Required(bool, default=False)


class Page(db.Entity):
    """ Page of an exam """
    path = Required(str)
    submission = Required(Submission)


class Problem(db.Entity):
    """ this will be initialized @ app initialization and immutable from then on. """
    name = Required(str)
    exam = Required(Exam)
    feedback_options = Set('FeedbackOption')
    solutions = Set('Solution')


class FeedbackOption(db.Entity):
    """ feedback option -- can be shared by multiple problems.
    this means non-duplicate rows for things like 'all correct',
    but means that care must be taken when "updating" and "deleting"
    options from the UI (not yet supported)
    """
    problem = Required(Problem)
    text = Required(str)
    description = Optional(str)
    score = Optional(int)
    solutions = Set('Solution')


class Solution(db.Entity):
    """ solution to a single problem """
    submission = Required(Submission)
    problem = Required(Problem)
    PrimaryKey(submission, problem)  # enforce uniqueness on this pair
    graded_by = Optional(Grader)  # if null, this has not yet been graded
    graded_at = Optional(datetime)
    image_path = Required(str)
    feedback = Set(FeedbackOption)
    remarks = Optional(str)

